/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritkCudaRegularMemoryImportImageContainer.txx
	Version:		$Revision: 2883 $
	Date:			$Date: 2012-04-02 17:01:22 +0200 (Mo, 02 Apr 2012) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#include "ritkCudaRegularMemoryImportImageContainer.h"
#include "ritkCudaArrayImportImageContainer.h"
#include "ritkCudaPitchedMemoryImportImageContainer.h"

#include "ritkDebugManager.h"

namespace ritk
{

//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>
::CudaRegularMemoryImportImageContainer()
{
	m_TmpCudaContainer = NULL;
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>
::~CudaRegularMemoryImportImageContainer()
{
	this->DeallocateManagedMemory();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
PrintSelf(std::ostream& os, itk::Indent indent) const
{
	os << "CudaRegularMemoryImportImageContainer" << std::endl;
	os << indent << "Memory @" << ((void*)m_CudaMemory.front()) << std::endl;
	Superclass::PrintSelf(os,indent);
} 


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
typename CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::Element*
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
AllocateElements(ElementIdentifier size) const
{
	// Sanity checks. This ensures that we can operate on the internal container size parameters
	if ( this->m_ContainerSize[0] == 0 )
	{
		LOG_DEB("Invalid size detected: did you forget to set them via SetSize()? RequestedSize="<<size);
		itkExceptionMacro(<<"Invalid size: did you forget to set them via SetSize()? RequestedSize="<<size);
	}
	ElementIdentifier Num = 1;
	for ( unsigned int dim = 0; dim < Self::Dimension; dim++ )
		Num *= this->m_ContainerSize[dim];
	if ( Num != size )
	{
		LOG_DEB("Mismatch between container size and requested size detected: " << Num << " vs. " << size);
		itkExceptionMacro(<<"Mismatch between widht/height and requested size detected: " << Num << " vs. " << size);
	}

	// Execute in the global CUDA context. Note that we can't use the size argument directly. Instead, we use the internal 2-D size!
	CudaBlockedConstMethodCallMacro(AllocateCudaMemory);

	// Call the superclass' implementation. This also returns the correct host memory pointer
	return Superclass::AllocateElements(size);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
DeallocateManagedMemory()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(DeallocateCudaMemory);

	// Call the superclass' implementation to deallocate host memory
	Superclass::DeallocateManagedMemory();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
SynchronizeDevice()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(CopyHostToDevice);

	// Synchronize MTimes
	Superclass::SynchronizeDevice();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
SynchronizeHost()
{
	// Execute in the global CUDA context
	CudaBlockedMethodCallMacro(CopyDeviceToHost);

	//std::cout << "CudaRegularMemoryImportImageContainer@" << this << ": synchronizing host..." << std::endl;
	
	// Synchronize MTimes
	Superclass::SynchronizeHost();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
DeepCopy(const ImportImageContainerType *Container)
{
	// Cast and pass
	ImportImageContainerType *NonConstContainer = const_cast<ImportImageContainerType*>(Container);
	if ( !NonConstContainer )
	{
		LOG_DEB("Could not remove const qualifier!");
		itkExceptionMacro(<<"Could not remove const qualifier!");
	}
	DeepCopy(NonConstContainer);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
DeepCopy(ImportImageContainerType *Container)
{
	// Call the superclass implementation first. This will ensure correct memory allocation and host memory copy.
	Superclass::DeepCopy(Container);

	// Check if we have a cuda container at all. If no, the GenericImportImageContainer should call a Synchronize() triggered by Superclass::DeepCopy() on us
	Superclass *CudaContainer = dynamic_cast<Superclass*>(Container);
	if ( !CudaContainer )
		return;

	// Check for MTimes: we do not need to copy device memory that is out-of-date. Eventuall, this has to be resolved by a synch!
	if ( CudaContainer->GetDeviceMTime() < CudaContainer->GetHostMTime() )
	{
		GenericImportImageContainerType::m_DeviceMTime = CudaContainer->GetDeviceTimeStamp();
		return;
	}

	// Device copy
	m_TmpCudaContainer = CudaContainer;
	CudaBlockedMethodCallMacro(DeepCopyCudaMemory);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
CopyToMappedResource(void *d_ptr, size_t size) const
{
	ritkCudaSafeCall(cudaMemcpyAsync(d_ptr, (void*)this->GetCudaMemoryPointer(), GenericImportImageContainerType::Capacity()*sizeof(Element), cudaMemcpyDeviceToDevice));
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
typename CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::Element*
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
GetCudaMemoryPointer()
{
	GenericImportImageContainerType::DeviceModified();
	return m_CudaMemory.front();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
const typename CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::Element*
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
GetCudaMemoryPointer() const
{
	return m_CudaMemory.front();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
AllocateCudaMemory() const
{
	// Allocate. Note that a correct size is checked in AllocateElements()
	Element *Mem;
	ritkCudaSafeCall(cudaMalloc((void**)&Mem, Superclass::m_ContainerSize[0]*sizeof(Element)));

	// Save the allocated memory at the end of our internal list such that it won't be deleted through the itk (de)allocate scheme
	m_CudaMemory.push_back(Mem);
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
DeallocateCudaMemory()
{
	// Sanity check
	if ( m_CudaMemory.size() == 0 )
	{
		return;
	}

	// Delete the first element in our internal list
	Element *Memory = m_CudaMemory.front();
	ritkCudaSafeCall(cudaFree((void*)Memory));
	m_CudaMemory.pop_front();
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
CopyDeviceToHost()
{
	// Copy data from device to host
	ritkCudaSafeCall(cudaMemcpy(GenericImportImageContainerType::GetBufferPointer(), this->GetCudaMemoryPointer(), GenericImportImageContainerType::Capacity()*sizeof(Element), cudaMemcpyDeviceToHost));
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
CopyHostToDevice()
{
	// Copy data from host to device
	ritkCudaSafeCall(cudaMemcpyAsync(this->GetCudaMemoryPointer(), GenericImportImageContainerType::GetBufferPointer(), GenericImportImageContainerType::Capacity()*sizeof(Element), cudaMemcpyHostToDevice));
}


//----------------------------------------------------------------------------
template<typename TElementIdentifier, typename TElement, unsigned int VDimension>
void
CudaRegularMemoryImportImageContainer<TElementIdentifier, TElement, VDimension>::
DeepCopyCudaMemory()
{
	// Regular memory -> Regular memory
	const CudaRegularMemoryImportImageContainer *RegularContainer = dynamic_cast<const CudaRegularMemoryImportImageContainer*>(m_TmpCudaContainer);
	if ( RegularContainer )
	{
		// Copy
		ritkCudaSafeCall(cudaMemcpyAsync
		(
			(void*)this->GetCudaMemoryPointer(),
			(const void*)RegularContainer->GetCudaMemoryPointer(),
			this->Capacity()*sizeof(Element),
			cudaMemcpyDeviceToDevice
		));

		// Update the timestamps. Take care if host and device are already synchronized!
		GenericImportImageContainerType::DeviceModified();
		if ( m_TmpCudaContainer->GetDeviceTimeStamp() == m_TmpCudaContainer->GetHostTimeStamp() )
			GenericImportImageContainerType::m_HostMTime = GenericImportImageContainerType::m_DeviceMTime;

		// Clean up 
		m_TmpCudaContainer = NULL;
		return;
	}

	// Array -> Regular memory
	typedef CudaArrayImportImageContainer<ElementIdentifier, Element> CudaArrayImportImageContainerType;
	const CudaArrayImportImageContainerType *ArrayContainer = dynamic_cast<const CudaArrayImportImageContainerType*>(m_TmpCudaContainer);
	if ( ArrayContainer )
	{
		// Copy
		ritkCudaSafeCall(cudaMemcpyFromArrayAsync
		(
			(void*)this->GetCudaMemoryPointer(),
			ArrayContainer->GetCudaMemoryArray(),
			0,
			0,
			GenericImportImageContainerType::Capacity()*sizeof(Element),
			cudaMemcpyDeviceToDevice
		));

		// Update the timestamps. Take care if host and device are already synchronized!
		GenericImportImageContainerType::DeviceModified();
		if ( m_TmpCudaContainer->GetDeviceTimeStamp() == m_TmpCudaContainer->GetHostTimeStamp() )
			GenericImportImageContainerType::m_HostMTime = GenericImportImageContainerType::m_DeviceMTime;

		// Clean up
		m_TmpCudaContainer = NULL;
		return;
	}

	// Pitched memory -> Regular memory
	typedef CudaPitchedMemoryImportImageContainer<ElementIdentifier, Element> CudaPitchedMemoryImportImageContainerType;
	const CudaPitchedMemoryImportImageContainerType *PitchedContainer = dynamic_cast<const CudaPitchedMemoryImportImageContainerType*>(m_TmpCudaContainer);
	if ( PitchedContainer )
	{
		// Get pitch parameters
		size_t Pitch = PitchedContainer->GetPitch();
		size_t Width = PitchedContainer->GetContainerSize()[0];
		size_t Height = PitchedContainer->GetContainerSize()[1];

		// Copy
		ritkCudaSafeCall(cudaMemcpy2DAsync
		(
			(void*)this->GetCudaMemoryPointer(),
			Width*sizeof(Element),
			(const void*)PitchedContainer->GetCudaMemoryPointer(),
			Pitch,
			Width*sizeof(Element),
			Height,
			cudaMemcpyDeviceToDevice
		));

		// Update the timestamps. Take care if host and device are already synchronized!
		GenericImportImageContainerType::DeviceModified();
		if ( m_TmpCudaContainer->GetDeviceTimeStamp() == m_TmpCudaContainer->GetHostTimeStamp() )
			GenericImportImageContainerType::m_HostMTime = GenericImportImageContainerType::m_DeviceMTime;

		// Clean up
		m_TmpCudaContainer = NULL;
		return;
	}

	// We can't handle this container
	LOG_DEB("Invalid container detected: " << m_TmpCudaContainer);
	itkExceptionMacro(<<"Invalid container detected: " << m_TmpCudaContainer);
}


} // ritk
