#include "KinectOpenNISourcePlugin.h"
#include <QTimer>

#include "ritkDebugManager.h"
#include "ritkRImageIOManager.h"

KinectOpenNISourcePlugin::KinectOpenNISourcePlugin()
{
	// Create source
	m_Source = KinectOpenNIRImageSourceType::New();

	// Initialize timer
	m_Timer = new QTimer();		
	m_Timer->stop();
	//connect(m_Timer, SIGNAL(timeout()), this, SLOT(RequestFrame()));
 
	// Create GUI and connect player related buttons
	m_Widget = new KinectOpenNISourcePluginWidget();
	
	// Connect elements from the widget
	connect(m_Widget->m_ConnectDisconnectPushButton,	SIGNAL(clicked()),	this,	SLOT(ConnectDisconnect())	);
	connect(m_Widget->m_PushButtonPause,				SIGNAL(clicked()),	this,	SLOT(Pause())				);
	connect(m_Widget->m_PushButtonPlayForward,			SIGNAL(clicked()),	this,	SLOT(Play())				);
	connect(m_Widget->m_PushButtonStepForward,			SIGNAL(clicked()),	this,	SLOT(RequestFrame())		);

	// Connect the UpdateGUI mechanism
	connect(this, SIGNAL(UpdateGUI()), m_Widget, SLOT(UpdateGUI()));

	// Default state
	SetState(Disconnected);
}


KinectOpenNISourcePlugin::~KinectOpenNISourcePlugin()
{
	// Clean up widget (this is mandatory!)
	if ( m_Widget )
		delete m_Widget;

	// Delete timer
	if ( m_Timer )
		delete m_Timer;

	// Disconnect reader
	if ( m_State == Playing )
		m_Source->Disconnect();
}


//----------------------------------------------------------------------------
QWidget*
KinectOpenNISourcePlugin::GetPluginGUI()
{
	return m_Widget;
}


//----------------------------------------------------------------------------
KinectOpenNISourcePlugin::SourcePlugin::SourcePointer 
KinectOpenNISourcePlugin::GetSource()
{	
	return (SourcePlugin::SourceType*)m_Source;
}


//----------------------------------------------------------------------------
void
KinectOpenNISourcePlugin::SetState(StateType State)
{
	// Lock
	{
		QMutexLocker locker(&m_StateMutex);
		m_State = State;

		// Update widget
		if ( m_Widget )
		{
			switch ( m_State )
			{
			case Disconnected:
				//std::cout << "Setting state to disconnected..." << std::endl;
				m_Widget->m_PushButtonPause->setEnabled(false);
				m_Widget->m_PushButtonPlayForward->setEnabled(false);
				m_Widget->m_PushButtonStepForward->setEnabled(false);
				m_Widget->m_ConnectDisconnectPushButton->setText(tr("Connect"));
				break;
			case Connected:
				//std::cout << "Setting state to connected..." << std::endl;
				m_Widget->m_PushButtonPause->setEnabled(false);
				m_Widget->m_PushButtonPlayForward->setEnabled(true);
				m_Widget->m_PushButtonStepForward->setEnabled(true);
				m_Widget->m_ConnectDisconnectPushButton->setText(tr("Disconnect"));
				break;
			case Playing:
				//std::cout << "Setting state to playing..." << std::endl;
				m_Widget->m_PushButtonPause->setEnabled(true);
				m_Widget->m_PushButtonPlayForward->setEnabled(false);
				m_Widget->m_PushButtonStepForward->setEnabled(false);
				break;
			}
			emit UpdateGUI();
		}
	}
}


//----------------------------------------------------------------------------
void
KinectOpenNISourcePlugin::ConnectDisconnect()
{
	// If we're connected we have to disconnect and vice versa
	if ( m_State == Connected || m_State == Playing )
	{
		if ( m_State == Playing )
			m_Timer->stop();
		m_Source->Disconnect();
		SetState(Disconnected);
	}
	else
	{
		try
		{
			m_Source->Connect();
		}
		catch ( itk::ExceptionObject &e )
		{
			std::cerr << e << std::endl;
		}
		SetState(Connected);
	}
}


//----------------------------------------------------------------------------
void
KinectOpenNISourcePlugin::Play()
{
	SetState(Playing);
	//m_Timer->start(25);

	// Connecting to the pipeline's new frame signal will automaticall trigger a frame request
	connect(m_AcquisitionPipeline, SIGNAL(NewFrame()), this, SLOT(RequestFrame()) );

	// One initial request mandatory
	RequestFrame();
}


//----------------------------------------------------------------------------
void
KinectOpenNISourcePlugin::Pause()
{
	//m_Timer->stop();
	disconnect(m_AcquisitionPipeline, SIGNAL(NewFrame()), this, SLOT(RequestFrame()) );
	SetState(Connected);
}


//----------------------------------------------------------------------------
void
KinectOpenNISourcePlugin::RequestFrame()
{
	// ======================= //
	// Update your source here //
	// ======================= //
	// ...

	m_AcquisitionPipeline->Lock();

	// We need to modify the source
	m_Source->Modified();

	m_AcquisitionPipeline->Unlock();

	// Trigger pipeline execution
	m_AcquisitionPipeline->ExecuteOnce();
}
