#ifndef KinectPoseImage_H__
#define KinectPoseImage_H__

#include "KinectImage.h"


/**	@class		KinectPoseImage
 *	@author		Sebastian Bauer
 *	@brief		Image for the Microsoft Kinect w/ user interface
 *
 *	@details
*/
template<class TReal, unsigned int NDimension=2>
class KinectPoseImage : public KinectImage<TReal,2>
{
public:
	/** @name Standard typedefs */
	//@{
	typedef KinectPoseImage						Self;
	typedef RImage<TReal,NDimension>			Superclass;
	typedef itk::SmartPointer<Self>				Pointer;
	typedef itk::SmartPointer<const Self>		ConstPointer;
	//@}

	/// Object creation
	itkNewMacro(Self);

	/// RTTI
	itkTypeMacro(KinectPoseImage, KinectImage);

	/// Image dimension
	itkStaticConstMacro(ImageDimension, unsigned int, Superclass::ImageDimension);

	/** @name Convenience methods to set the LargestPossibleRegion, BufferedRegion and RequestedRegion. Allocate must still be called.*/
	//@{
	virtual void SetRegions(RegionType region);
	virtual void SetRegions(SizeType size);
	//@}

	/**	@name LargestPossibleRegion, BufferedRegion, RequestedRegion. Calling these methods will correctly set the SX and SY camera parameters!*/
	//@{
	virtual void SetLargestPossibleRegion(const RegionType &region);
	virtual void SetRequestedRegion(const RegionType &region);
	virtual void SetBufferedRegion(const RegionType &region);
	//@}

	/// Allocate the internal images (inherited from itk::ImageBase)
	virtual void Allocate();

	/** @name Set the spacing. Will be delegated to the internal images*/
	//@{
	virtual void SetSpacing(const SpacingType &Spacing);
	virtual void SetSpacing(const double Spacing[ImageDimension]);
	virtual void SetSpacing(const float Spacing[ImageDimension]);
	//@}

	/** @name Set the origin. Will be delegated to the internal images*/
	//@{
	virtual void SetOrigin(const PointType &Origin);
	virtual void SetOrigin(const double Origin[ImageDimension]);
	virtual void SetOrigin(const float Origin[ImageDimension]);
	//@}

	/// Set the direction 
	virtual void SetDirection(const DirectionType direction);

	/// Copy information from the specified data set
	virtual void CopyInformation(const DataObject *data);

	/**	@brief	Graft from another image
	 *	@param	data	The object to graft
	 *
	 *	@details
	 *	This method will graft the data and information from another object.
	 *	If the object can be casted to a KinectPoseImage the RGBImage will be grafted, too.
	 */
	virtual void Graft(const DataObject *data);

	/**	@brief	Perform a deep copy
	 *	@param	data	The data to copy from
	 *
	 *	@details
	 *	This method will perform a deep copy (in contrast to the Graft() 
	 *	method that performs a shallow copy) from the given object. 
	 *	If the object can be casted to a KinectPoseImage the RGBImage will be grafted, too.
	 */
	virtual void DeepCopy(const DataObject *data);

	/// Disconnect the image from the pipeline
	virtual void DisconnectPipeline();

	/// Set parameters used to convert from projective to real world space
	void SetConversionParameters(float XToZ, float YToZ);
	void GetConversionParameters(float *XToZ, float *YToZ);

	/// Compute the world coord image
	virtual void ComputeRangeToWorldCoords();

	xn::DepthGenerator *m_UserGenerator;


protected:
	/// Constructor
	KinectPoseImage();
	/// Destructor
	~KinectPoseImage();

	/// Print the object on a stream
	void PrintSelf(std::ostream& os, itk::Indent indent) const;

protected:
	/**	@name Members used to compute the world coords out of the range data */
	//@{
	TReal m_XToZ;
	TReal m_YToZ;
	//@}

	/**	@name	The range to world coord filter */
	//@{
	typedef KinectRangeToWorldFilter<RangeImageType,FlagImageType,WorldCoordImageType> R2WCFilterType;
	typename R2WCFilterType::Pointer m_R2WCFilter;
	//@}

private:
	/// purposely not implemented
	KinectPoseImage(Self&); 
	/// purposely not implemented
	void operator=(const Self&); 
};

/** @name Template instantiations */
//@{
typedef class __declspec(dllexport) KinectPoseImage<float> KinectPoseImageF2;
//@}


//#include "KinectPoseImage.txx"

#endif KinectPoseImage_H__